@@ -2,13 +2,13 @@
 
   readline.c - GNU Readline module
 
-  $Author$
+  $Author: nobu $
   created at: Wed Jan 20 13:59:32 JST 1999
 
   Copyright (C) 1997-2008  Shugo Maeda
   Copyright (C) 2008-2013  Kouji Takao
 
-  $Id$
+  $Id: readline.c 52453 2015-11-05 01:09:17Z nobu $
 
   Contact:
    - Kouji Takao <kouji dot takao at gmail dot com> (current maintainer)
@@ -229,8 +229,7 @@ readline_getc(FILE *input)
                 goto again;
             rb_sys_fail("rb_wait_for_single_fd");
         }
-        errno = data.err;
-        rb_sys_fail("read");
+        rb_syserr_fail(data.err, "read");
     }
     return data.ret;
 }
@@ -557,8 +556,7 @@ readline_s_set_input(VALUE self, VALUE input)
         if (f == NULL) {
             int save_errno = errno;
             close(fd);
-            errno = save_errno;
-            rb_sys_fail("fdopen");
+            rb_syserr_fail(save_errno, "fdopen");
         }
         rl_instream = readline_rl_instream = f;
         readline_instream = input;
@@ -594,8 +592,7 @@ readline_s_set_output(VALUE self, VALUE output)
         if (f == NULL) {
             int save_errno = errno;
             close(fd);
-            errno = save_errno;
-            rb_sys_fail("fdopen");
+            rb_syserr_fail(save_errno, "fdopen");
         }
         rl_outstream = readline_rl_outstream = f;
         readline_outstream = output;
@@ -966,7 +963,7 @@ readline_attempted_completion_function(const char *text, int start, int end)
     enc = rb_locale_encoding();
     encobj = rb_enc_from_encoding(enc);
     for (i = 0; i < matches; i++) {
-        temp = rb_obj_as_string(RARRAY_PTR(ary)[i]);
+        temp = rb_obj_as_string(RARRAY_AREF(ary, i));
         StringValueCStr(temp);  /* must be NUL-terminated */
         rb_enc_check(encobj, temp);
         result[i + 1] = (char*)malloc(RSTRING_LEN(temp) + 1);
